import warnings
import numpy as np

class Noises:
    @staticmethod
    def uniform(low = 0, high = 0.1):
        return lambda value: value + np.random.uniform(low = low, high = high)

    @staticmethod
    def normal(center = 0, sd = 0.1):
        return lambda value: value + np.random.normal(loc = center, scale = sd)



class Transformation:

    def __init__(self, transformed_function, shift_step = None, rotation_matrix = None, noise_generator = None, seed = None):
        """
        Creates Transformation object

        Parameters
        ----------
        transformed_function : function or class callable object
            transformed function.
        shift_step : numpy 1D array/None, optional
            array of shifts by each dimension or None. The default is None.
        rotation_matrix : 2D-array/int/None, optional
            2D ortogonal rotation matrix or dimension for creating random rotation matrix or None if no rotate. The default is None.
        noise_generator : function, optional
            function gets current value and returns value with some noise. The default is None.
        seed : int, optional
            random seed for rotation matrix if needed reproduce. The default is None.

        """
        if not (seed is None):
            np.random.seed(seed)

        self.is_noised = not (noise_generator is None)
        self.is_rotated = not (rotation_matrix is None)
        self.is_shifted = not (shift_step is None)

        self.bounds = transformed_function.bounds
        if self.is_shifted:
            xmin, xmax, ymin, ymax = self.bounds
            self.bounds = xmin + shift_step[0], xmax + shift_step[0], ymin + shift_step[1], ymax + shift_step[1]
        #raise Exception()
        if (shift_step is None) and (rotation_matrix is None) and (noise_generator) is None:
            warnings.warn("No sense of transformation when all preparations are None!")
            self.f = lambda arr: transformed_function(arr)

            return
        
        empty_func = lambda arr: arr

        if self.is_shifted:
            
            assert (type(shift_step) == np.ndarray), "shift_step must be numpy array or None!"

            self.shifter = lambda arr: arr - shift_step
            self.unshifter = lambda arr: arr + shift_step
        else:
            self.shifter = empty_func
            self.unshifter = empty_func
        

        if self.is_rotated:
            if type(rotation_matrix) == np.ndarray:

                assert (np.allclose(rotation_matrix.T, np.linalg.inv(rotation_matrix))), f"Rotation matrix must be ortogonal!"

                # create rotator
                self.rotator = lambda arr: arr.dot(rotation_matrix)
            else:
                assert (type(rotation_matrix) == int), "rotation_matrix is not int dim and not a matrix!"
                # init rotator
                rotation_matrix, _ = np.linalg.qr(np.random.random((rotation_matrix, rotation_matrix)), mode='complete')
                self.rotator = lambda arr: arr.dot(rotation_matrix)
                self.unrotator = lambda arr: arr.dot(rotation_matrix.T)
        else:
            self.rotator = empty_func
            self.unrotator = empty_func
            self.is_rotated = False

        
        self.noiser = (lambda val: val) if noise_generator is None else noise_generator

        self.f = lambda arr: self.noiser(transformed_function(self.rotator(self.shifter(arr))))

        self.x_best = None
        self.f_best = None

        if not self.is_noised and hasattr(transformed_function, 'x_best'):
            if not (transformed_function.x_best is None):
                self.x_best = self.unrotator(self.unshifter(transformed_function.x_best))
                self.f_best = transformed_function(self.x_best)


    def __call__(self, arr):
        return self.f(arr)












from mpl_toolkits.mplot3d import Axes3D  
import numpy as np 
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator, LinearLocator, FormatStrFormatter

from matplotlib import cm

from OppOpPopInit import OppositionOperators


def get_good_arrow_place(optimum, bounds):
    opt = np.array(optimum)
    minimums = np.array([bounds[0], bounds[2]])
    maximums = np.array([bounds[1], bounds[3]])

    t1 = OppositionOperators.Continual.over(minimums, maximums)(opt)
    t2 = OppositionOperators.Continual.quasi_reflect(minimums, maximums)(t1)

    return tuple(t2)



def plot_3d(func, points_by_dim = 50, title = '', bounds = None, show_best_if_exists = True, save_as = None, cmap = 'twilight', plot_surface = True, plot_heatmap = True):
    """
    Plots function surface and/or heatmap

    Parameters
    ----------
    func : class callable object
        Object which can be called as function.
    points_by_dim : int, optional
        points for each dimension of plotting (50x50, 100x100...). The default is 50.
    title : str, optional
        title of plot with LaTeX notation. The default is ''.
    bounds : tuple, optional
        space bounds with structure (xmin, xmax, ymin, ymax). The default is None.
    show_best_if_exists : boolean, optional
        point best solution by arrow if x_best exists. The default is True.
    save_as : str/None, optional
        file path to save image (None if not needed). The default is None.
    cmap : str, optional
        color map of plot. The default is 'twilight'.
    plot_surface : boolean, optional
        plot 3D surface. The default is True.
    plot_heatmap : boolean, optional
        plot 2D heatmap. The default is True.
    """
    
    assert (plot_surface or plot_heatmap), "should be plotted at least surface or heatmap!"

    if bounds is None:
        bounds = func.bounds
    
    xmin, xmax, ymin, ymax = bounds

    x = np.linspace(xmin, xmax, points_by_dim)
    y = np.linspace(ymin, ymax, points_by_dim)


    a, b = np.meshgrid(x, y)
    
    data = np.empty((points_by_dim, points_by_dim))
    for i in range(points_by_dim):
        for j in range(points_by_dim):
            data[i,j] = func(np.array([x[i], y[j]]))
    
    a = a.T
    b = b.T

    l_a, r_a, l_b, r_b = xmin, xmax, ymin, ymax
    
    l_c, r_c = data.min(), data.max()

    levels = MaxNLocator(nbins=15).tick_values(l_c,r_c)

    if plot_heatmap and plot_surface:

        fig = plt.figure(figsize=(16, 6))
        ax1 = fig.add_subplot(1,2,1)
        ax2 = fig.add_subplot(1,2,2, projection='3d')
    else:
        fig = plt.figure()
        if plot_heatmap:
            ax1 = fig.gca()
        else:
            ax2 = fig.gca(projection='3d')

    title = r"$\bf{" + title+ r"}$"
    min_title = title[::]

    def base_plot():
        c = ax1.contourf(a, b, data , cmap=cmap, levels = levels, vmin=l_c, vmax=r_c)       
        name = title
        ax1.set_title( name, fontsize = 15)
        ax1.axis([l_a, r_a, l_b, r_b])
        fig.colorbar(c)

    if plot_surface:
        # Plot the surface.
        surf = ax2.plot_surface(a, b, data, cmap =  cmap,  linewidth=0, antialiased=False)

        # Customize the z axis.
        ax2.set_xlabel('first dim', fontsize=10)
        ax2.set_ylabel('second dim', fontsize=10)
        ax2.set_zlim(l_c, r_c)
        
        ax2.zaxis.set_major_locator(LinearLocator(4))
        #ax2.zaxis.set_major_formatter(FormatStrFormatter('%.2f'))

        # Add a color bar which maps values to colors.
        if not plot_heatmap: fig.colorbar(surf)#, shrink=0.5, aspect=5)
        
        ax2.contour(a, b, data, zdir='z', offset=0, cmap =  cmap)
        ax2.view_init(60, 35)
        ax2.set_title( min_title , fontsize = 15, loc = 'right')
    

    if not (func.x_best) is None:
        title += f"\n best solution: f{func.x_best} = {round(func(func.x_best))}"

        if show_best_if_exists and plot_heatmap:
            #xytext = tuple(np.array([(xmax+xmin)/2, (ymax+ymin)/2]) + np.random.uniform(-func.x_best/2, func.x_best/2, 2)) #tuple(np.random.uniform(0, min((xmax+xmin)/2, (ymax+ymin)/2), 2))

            xytext = get_good_arrow_place(func.x_best, bounds)

            bbox = dict(boxstyle ="round", fc ="0.8") 
            arrowprops = dict( 
                arrowstyle = "->", 
                connectionstyle = "angle, angleA = 0, angleB = 90, rad = 10")
    
            ax1.annotate(f'global minimum', xy= tuple(func.x_best), xytext = xytext,
                arrowprops=dict(facecolor='red', shrink=0.05), 
                #color = 'red',
                bbox = bbox#, arrowprops = arrowprops
                )

    if plot_heatmap: base_plot()


    fig.tight_layout()

    if save_as != None:
        plt.savefig(save_as, dpi = 250)
    
    plt.show()

    plt.close()











from .functions import Sphere, Ackley, AckleyTest, Rosenbrock, Fletcher, Griewank, Penalty2, Quartic, Rastrigin, SchwefelDouble, SchwefelMax, SchwefelAbs, SchwefelSin, Stairs, Abs, Michalewicz, Scheffer, Eggholder, Weierstrass

from .transformations import Transformation, Noises

from .plot_func import plot_3d





#
#
# See also https://www.sfu.ca/~ssurjano/optimization.html
#
#



import math
import numpy as np


def easy_bounds(bound):
    return (-bound, bound, -bound, bound)

def check_dim(dim, min = 1):
    assert (type(dim) == int and dim >=min), f"Dimension should be int and not less than {min} for this function (got {dim})"


class Sphere:

    b = 5.12

    def __init__(self, dim, degree = 2):

        check_dim(dim, 1)

        self.deg = degree
        self.x_best = np.zeros(dim) if degree % 2 == 0 else None
        self.f_best = 0 if not (self.x_best is None) else None

        self.bounds = easy_bounds(Sphere.b)
    
    def __call__(self, vec):
        return sum((x**self.deg for x in vec))


class Ackley:
    
    b = 3

    def __init__(self, dim):

        check_dim(dim, 1)

        self.x_best = np.zeros(dim)
        self.f_best = 0

        self.bounds = easy_bounds(Ackley.b)

        self.bias = 20 + math.e
        self.pi2 = 2 * math.pi
    
    def __call__(self, vec):

        s1 = sum((x*x for x in vec))/vec.size
        s2 = sum((math.cos(self.pi2 * x) for x in vec))/vec.size


        return self.bias - 20*math.exp(-0.2*s1) - math.exp(s2)


class AckleyTest:
    
    b = 30

    def __init__(self, dim):

        check_dim(dim, 2)

        self.x_best = None
        self.f_best = None

        self.bounds = easy_bounds(AckleyTest.b)

        self.exp = math.exp(-0.2)
    
    def __call__(self, vec):

        s = sum((3*(math.cos(2*vec[i]) + math.sin(2*vec[i+1])) + self.exp * math.sqrt(vec[i]**2 + vec[i+1]**2) for i in range(vec.size-1)))

        return s


class Rosenbrock:
    
    b = 2.048

    def __init__(self, dim):

        check_dim(dim, 2)

        self.x_best = np.ones(dim)
        self.f_best = 0

        self.bounds = easy_bounds(Rosenbrock.b)

    
    def __call__(self, vec):

        s = sum(( 100 * (vec[i+1] - vec[i]**2) ** 2 + (vec[i] - 1)**2 for i in range(vec.size-1)))

        return s

class Fletcher:
    
    b = math.pi

    def __init__(self, dim, seed = None):

        if seed != None:
            np.random.seed(seed)

        check_dim(dim, 1)

        self.x_best = np.random.uniform(-np.pi, np.pi, dim)
        self.f_best = 0
        self.bounds = easy_bounds(Fletcher.b)

        self.a = np.random.uniform(-100, 100, (dim, dim))
        self.b = np.random.uniform(-100, 100, (dim, dim))

        self.A = np.sum(self.a * np.sin(self.x_best) + self.b * np.cos(self.x_best), axis = 0)


        

    
    def __call__(self, vec):

        B = np.sum(self.a * np.sin(vec) + self.b * np.cos(vec) ,axis = 0)
        #raise Exception()
        return  sum((a-b)**2 for a, b in zip(self.A, B))


class Griewank:
    
    b = 600

    def __init__(self, dim):

        check_dim(dim, 1)

        self.x_best = np.zeros(dim)
        self.f_best = 0

        self.bounds = easy_bounds(Griewank.b)

    
    def __call__(self, vec):

        s = sum(( x*x for x in vec))/4000
        p = math.prod((math.cos(x/math.sqrt(i+1)) for i, x in enumerate(vec)))

        return 1 + s - p



class Penalty2:
    
    b = 50

    def __init__(self, dim, a=5, k=100, m=4):

        check_dim(dim, 2)

        self.x_best = np.ones(dim)
        self.f_best = 0

        self.bounds = easy_bounds(Penalty2.b)

        self.a, self.k, self.m = a, k, m
        self.pi2 = 2 * math.pi
        self.pi3 = 3 * math.pi

    
    def __call__(self, vec):

        a, k, m = self.a, self.k, self.m

        u = np.sum((vec[vec>a]-a)**m) + np.sum((-vec[vec<-a]-a)**m)

        s1 = 10 * math.sin(self.pi3*vec[0])**2 + (vec[-1]-1)**2 * (1 + math.sin(self.pi2 * vec[-1]**2))

        s2 = sum(((vec[i]-1)**2 * (1 + math.sin(self.pi3 * vec[i+1]**2)) for i in range(vec.size-1)))

        return k*u + 0.1 * (s1 + s2)


class Quartic:
    
    b = 1.28

    def __init__(self, dim):

        check_dim(dim, 1)

        self.x_best = np.zeros(dim)
        self.f_best = 0

        self.bounds = easy_bounds(Quartic.b)

    
    def __call__(self, vec):

        s = sum(( (i+1)*x**4 for i, x in enumerate(vec)))

        return s



class Rastrigin:
    
    b = 5.12

    def __init__(self, dim):

        check_dim(dim, 1)

        self.x_best = np.zeros(dim)
        self.f_best = 0

        self.bounds = easy_bounds(Rastrigin.b)


        self.pi2 = math.pi*2
        self.bias = 10*dim

    
    def __call__(self, vec):

        s = sum(( x*x - math.cos(self.pi2*x)*10 for x in vec))

        return self.bias + s


class SchwefelDouble:
    
    b = 65.536

    def __init__(self, dim):
        check_dim(dim)

        self.x_best = np.zeros(dim)
        self.f_best = 0

        self.bounds = easy_bounds(SchwefelDouble.b)


    
    def __call__(self, vec):

        cs = np.cumsum(vec)

        s = sum((cs[i]**2 for i in range(vec.size)))

        return s


class SchwefelMax:
    
    b = 100

    def __init__(self, dim):
        check_dim(dim)

        self.x_best = np.zeros(dim)
        self.f_best = 0

        self.bounds = easy_bounds(SchwefelMax.b)

    def __call__(self, vec):

        return np.abs(vec).max()

class SchwefelAbs:
    
    b = 10

    def __init__(self, dim):

        check_dim(dim)

        self.x_best = np.zeros(dim)
        self.f_best = 0

        self.bounds = easy_bounds(SchwefelAbs.b)

    def __call__(self, vec):

        mod = np.abs(vec)

        return np.sum(mod) + np.prod(mod)


class SchwefelSin:
    
    b = 500

    def __init__(self, dim):

        check_dim(dim, 1)

        self.x_best = np.full(dim, 420.9687)
        self.f_best = -12965.5

        self.bounds = easy_bounds(SchwefelSin.b)

    def __call__(self, vec):

        return -sum((x*math.sin(math.sqrt(abs(x))) for x in vec))


class Stairs:
    
    b = 6

    def __init__(self, dim):

        check_dim(dim)

        self.x_best = np.zeros(dim)
        self.f_best = 0

        self.bounds = easy_bounds(Stairs.b)

    def __call__(self, vec):

        return sum(( math.floor(x + 0.5)**2 for x in vec))


class Abs:
    
    b = 10

    def __init__(self, dim):

        check_dim(dim)

        self.x_best = np.zeros(dim)
        self.f_best = 0

        self.bounds = easy_bounds(Abs.b)

    def __call__(self, vec):

        return sum(( abs(x) for x in vec))


class Michalewicz:
    

    def __init__(self, m = 10):

        self.x_best = None
        self.f_best = None

        self.bounds = (0, math.pi, 0, math.pi)

        self.m = m*2

    def __call__(self, vec):

        return -sum(( math.sin(x)*math.sin((i+1)*x**2/math.pi)**self.m for i, x in enumerate(vec)))


class Scheffer:
    
    b = 7

    def __init__(self, dim):

        check_dim(dim, 2)

        self.x_best = np.zeros(dim)
        self.f_best = 0

        self.bounds = easy_bounds(Scheffer.b)


    def __call__(self, vec):

        return 0.5 + sum((  (math.sin( vec[i]**2 - vec[i+1]**2 )**2 - 0.5) / (1 + 0.001*(vec[i]**2 + vec[i+1]**2) )**2  for i in range(vec.size-1)))


class Eggholder:
    
    b = 512

    def __init__(self, dim):

        check_dim(dim, 2)

        self.x_best = None
        self.f_best = None

        self.bounds = easy_bounds(Eggholder.b)


    def __call__(self, vec):

        return -sum(( (vec[i+1] + 47) * math.sin(math.sqrt(abs(vec[i+1] + vec[i]/2 + 47))) + vec[i] * math.sin(math.sqrt(abs(vec[i]-vec[i+1] - 47))) for i in range(vec.size-1)))


class Weierstrass:
    
    b = 0.5

    def __init__(self, dim, a = 0.5, b = 3, kmax = 20):

        check_dim(dim, 1)

        self.x_best = np.zeros(dim)
        self.f_best = 0

        self.bounds = easy_bounds(Weierstrass.b)

        self.ak = np.array([a**k for k in range(kmax+1)])
        self.pibk = np.pi * np.array([b**k for k in range(kmax+1)])

        self.bias = -dim*np.sum(self.ak*np.cos(self.pibk))


    def __call__(self, vec):
    
        return self.bias + sum(( sum((ak * math.cos(x*pibk)  for ak, pibk in zip(self.ak, self.pibk) )) for x in vec*2 + 1 ))




if __name__ == '__main__':
    
    arr = np.array([0.1, 0.2, 0.3])
    
    dim = arr.size
    
    funcs = [
        Sphere(dim, degree = 2),
        Ackley(dim),
        AckleyTest(dim),
        Rosenbrock(dim),
        Fletcher(dim, seed = None),
        Griewank(dim),
        Penalty2(dim, a=5, k=100, m=4),
        Quartic(dim),
        Rastrigin(dim),
        SchwefelDouble(dim),
        SchwefelMax(dim),
        SchwefelAbs(dim),
        SchwefelSin(dim),
        Stairs(dim),
        Abs(dim),
        Michalewicz(m = 10),
        Scheffer(dim),
        Eggholder(dim),
        Weierstrass(dim, a = 0.5, b = 3, kmax = 20)
    ]

    

    for f in funcs:
        print(f(arr))





















import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="OptimizationTestFunctions", 
    version="1.0.0",
    author="Demetry Pascal",
    author_email="qtckpuhdsa@gmail.com",
    maintainer = ['Demetry Pascal'],
    description="PyPI package containing collection of optimization test functions and some useful methods for working with them",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PasaOpasen/OptimizationTestFunctions",
    keywords=['optimization', 'evolutionary algorithms', 'fast', 'easy', 'evolution', 'generator', 'test', 'test-functions', '3D', 'functions'],
    packages = setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]#,
    #install_requires=['numpy']
    
    )





